from __future__ import annotations


class CharaChorderException(Exception):
    """Base exception class for charachorder"""


class UnknownProduct(CharaChorderException):
    """An exception raised when the pid of the peripheral is foreign"""

    def __init__(self, id: int):
        super().__init__(
            f'Device with product id "{id}" cannot be parsed as a CharaChorder device'
        )


class UnknownVendor(CharaChorderException):
    """An exception raised when the vid of the peripheral is foreign"""

    def __init__(self, id: int):
        super().__init__(
            f'Device with vendor id "{id}" cannot be parsed as a CharaChorder device'
        )


class SerialException(CharaChorderException):
    """An exception raised when something went wrong during Serial I/O"""


class AutoReconnectFailure(SerialException):
    """An exception raised when auto-reconnect after a restart failed"""


class ReconnectTimeout(AutoReconnectFailure):
    """An exception raised when auto-reconnect failed due to a timeout"""

    def __init__(self) -> None:
        super().__init__(
            "Auto reconnect after the restart has failed. You will need to re-create this object to do further Serial I/O"
        )


class TooManyDevices(AutoReconnectFailure):
    """An exception raised when auto-reconnect failed due to ambiguity in devices"""

    def __init__(self) -> None:
        super().__init__(
            "It looks like you have two or more devices of the same model connected simultaneously. This causes the restart to not reconnect to the correct device. You will need to re-create this object to do further Serial I/O"
        )


class SerialConnectionNotFound(SerialException):
    def __init__(self) -> None:
        super().__init__(
            "Serial connection does not exist. Did you open the connection?"
        )


class UnknownCommand(SerialException):
    """An exception raised when an unknown command is passed to the Serial API"""

    def __init__(self, command: str):
        super().__init__(f'The command "{command}" does not exist')


class InvalidParameter(SerialException):
    """An exception raised when the given parameter code does not exist"""

    def __init__(self, code: str):
        super().__init__(f'The parameter "{code}" does not exist.')


class InvalidParameterInput(SerialException):
    """An exception raised when the given input for a parameter code is invalid"""

    def __init__(self, code: str, value: int | str):
        super().__init__(
            f'The parameter value "{value}" for the code "{code}" is invalid.'
        )



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Self


class Chord(tuple):
    def to_hex(self) -> str:
        hexadecimal = 0
        for i in range(1, len(self) + 1):
            hexadecimal |= (ord(self[-i]) & 0x3FF) << (12 - i) * 10
        return format(hexadecimal, "032X")

    @classmethod
    def from_hex(cls, hexadecimal: int | str) -> Self:
        if isinstance(hexadecimal, int):
            chord = hexadecimal
        elif isinstance(hexadecimal, str):
            chord = int(hexadecimal, 16)

        actions = []
        for _ in range(12):
            action_code = int(chord & 0x3FF)
            if action_code != 0:
                actions.append(chr(action_code))
            chord >>= 10

        return cls(actions)


class ChordPhrase(tuple):
    def to_hex(self) -> str:
        def compress_phrase(phrase):
            buffer = bytearray(len(phrase) * 2)
            i = 0
            for char in phrase:
                action = ord(char)
                if action > 0xFF:
                    buffer[i] = action >> 8
                    i += 1
                buffer[i] = action & 0xFF
                i += 1
            return buffer[:i]

        compressed_actions = compress_phrase(self)
        return "".join(format(action, "02X") for action in compressed_actions)

    @classmethod
    def from_hex(cls, hexadecimal: str) -> Self:
        numeric_action_codes = []
        for i in range(0, len(hexadecimal), 2):
            numeric_action_codes.append(int(hexadecimal[i : i + 2], 16))

        action_codes = []
        for i, action_code in enumerate(numeric_action_codes):
            if action_code in range(32):  # 10-bit scan code
                action_codes[i + 1] = (action_code << 8) | action_codes[i + 1]

            elif action_code in range(32, 127):  # Alphanumeric
                action_codes.append(chr(action_code))

            elif action_code == 296:  # Line break
                action_codes.append("\n")

            elif action_code == 298 and len(action_codes) > 0:  # Backspace
                action_codes.pop()

            elif action_code == 299:  # Tab
                action_codes.append("\t")

            elif action_code == 544:  # Spaceright
                action_codes.append(" ")

            elif action_code > 126:  # Currently unsupported
                action_codes.append(f"<{action_code}>")

            else:
                action_codes.append(chr(action_code))
        return cls(action_codes)


class KeymapCode(Enum):
    primary = 0xA1
    secondary = 0xA2
    tertiary = 0xA3


class OperatingSystem(Enum):
    Windows = 0x0
    MacOS = 0x1
    Linux = 0x2
    iOS = 0x3
    Android = 0x4
    Unknown = 0x255
